/* Copyright 1991 Digital Equipment Corporation.               */
/* All rights reserved. See copyright notice in attached file COPYRIGHT. */
/* Last modified on Mon Feb 18 15:47:04 1991 by burrows        */
#include <stdio.h>
#include <errno.h>

extern void exit ();
extern void free ();
extern char *malloc ();
extern void perror ();

static int getnext ();

struct input {
	char buf[512];
	long n;
	int c;
	FILE *fp;
	int same;
	int after;
	struct input *next;
	char *name;
};

struct output {
	char buf[512];
	long n;
	FILE *fp;
};

main (argc, argv)
int argc;
char *argv[];
{
	struct output output;
	struct input *ipl = 0;
	register struct input *ip;
	register struct output *op = &output;
	int usedstdin = 0;
	int argn;

	if (argc == 1) {
		(void) fprintf (stderr, "usage: msgmerge file [file ..]\n");
		exit (2);
	}

	ipl = 0;

	for (argn = 1; argn < argc; argn++) {
		ip = (struct input *) malloc (sizeof (struct input));
		if (ip == 0) {
			perror ("malloc");
			exit (2);
		}

		ip->buf[0] = '\0';
		ip->n = 0;
		ip->same = 0;
		ip->after = 0;
		ip->next = ipl;
		ip->name = argv[argn];

		if (strcmp (argv[argn], "-") == 0 && !usedstdin) {
			usedstdin = 1;
			ip->fp = stdin;
			ip->name = "stdin";
		} else {
			ip->fp = fopen (argv[argn], "r");
			if (ip->fp == 0) {
				perror (argv[argn]);
				exit (2);
			}
		}
		ip->c = getc (ip->fp);
		if (getnext (ip) < 0) {
			if (ferror (ip->fp)) {
				perror (ip->name);
				exit (2);
			}
			if (ip->fp != stdin)
				(void) fclose (ip->fp);
			free ((char *) ip);
		} else
			ipl = ip;
	}
			
	op->buf[0] = '\0';
	op->n = -1;
	op->fp = stdout;

	ip = ipl;

	while (ipl != 0) {
		{
			register struct input *rip = ipl;

			while (rip != 0) {
				if (rip != ip && (!rip->after || ip->same>=0)) {
					if (ip->same < 0 && rip->same < 0) {
						if (ip->n > rip->n) {
							ip->after = 0;
							ip = rip;
						} else
							rip->after = 0;
					} else {
						register char *p = ip->buf;
						register char *q = rip->buf;

						while (*p == *q && *p != '\0')
							p++, q++;
						if (*p < *q)
							rip->after = 1;
						else if (*p > *q) {
							ip->after = 1;
							ip = rip;
						} else if (ip->n > rip->n) {
							ip->after = 0;
							ip = rip;
						} else
							rip->after = 0;
					}
				}

				rip = rip->next;
			}

			ip->after = 0;
		}

		{
			register FILE *fp = op->fp;
			register long n = ip->n;

			if (ip->same >= 0) {
				register char *p = op->buf+ip->same;
				register char *q = ip->buf+ip->same;

				while (*p == *q && *p != '\0')
					p++, q++;
			    
				if (*q != *p) {
					if (p - op->buf > 0x3f) {
						p = &op->buf[0x3f];
						q = &ip->buf[0x3f];
					}
					putc (0x80 | (p - op->buf), fp);
					while (*q != '\0') {
						putc (*q, fp);
						*p++ = *q++;
					}
					putc ('\0', fp);
					*p = '\0';
					op->n = -1;
				}
			}

			if (n != op->n) {
				if (op->n == -1) {
					op->n = n;
				} else {
					n -= op->n;
					op->n += n;
				}
				while (n > 0x7f) {
					putc (0xc0 | n, fp);
					n >>= 6;
				}
				putc (n, fp);
			}
		}

		{
			register int c = ip->c;
			register FILE *fp = ip->fp;

			if ((c & 0xc0) == 0x80) {
				register char *s;

				c &= 0x3f;
				s = &ip->buf[c];
				ip->same = c;

				c = getc (fp);
				
				while (c != '\0' && c != EOF) {
					*s++ = c;
					c = getc (fp);
				}
				if (c != EOF)
					c = getc (fp);

				*s = '\0';

				ip->n = 0;
				ip->c = c;
			} else
				ip->same = -1;
			
			if ((c & 0x180) == 0) {
				ip->n += c;
				ip->c = getc (fp);
			} else if ((c & 0x1c0) == 0xc0) { 
				register long n = (c & 0x3f);
				register int shift = 6;

				c = getc (fp);
				while ((c & 0x1c0) == 0xc0) {
					n |= (c & 0x3f) << shift;
					shift += 6;
					c = getc (fp);
				}
				n |= c << shift;
				ip->n += n;
				ip->c = getc (fp);
			} else {
				register struct input **pip = &ipl;

				if (ferror (ip->fp)) {
					perror (ip->name);
					exit (2);
				}
				if (ip->fp != stdin)
					(void) fclose (ip->fp);
				while (*pip != 0) {
					if (*pip == ip) {
						*pip = ip->next;
						break;
					}
					pip = &((*pip)->next);
				}
				free ((char *) ip);
				ip = ipl;
			}
		}
	}

	(void) fflush (stdout);

	if (ferror (stdout) || (fsync (fileno (stdout)) < 0 && errno!=EINVAL)) {
		perror ("stdout");
		exit (2);
	}

	exit (0);
}

static int getnext (ip)
register struct input *ip;
{
	register int c = ip->c;
	register FILE *fp = ip->fp;

	if ((c & 0xc0) == 0x80) {
		register char *s;

		c &= 0x3f;
		s = &ip->buf[c];
		ip->same = c;

		c = getc (fp);
		
		while (c != '\0' && c != EOF) {
			*s++ = c;
			c = getc (fp);
		}
		if (c != EOF)
			c = getc (fp);

		*s = '\0';

		ip->n = 0;
		ip->c = c;
	} else
		ip->same = -1;

	if ((c & 0xc0) != 0x80 && c != EOF) { 
		register long n = 0;
		register int shift = 0;

		while ((c & 0xc0) == 0xc0 && c != EOF) {
			n |= (c & 0x3f) << shift;
			shift += 6;
			c = getc (fp);
		}
		n |= (c & 0x7f) << shift;
		n += ip->n;
		ip->n = n;
		ip->c = getc (fp);
		return (0);
	}
	return (-1);
}
