/* Copyright 1991 Digital Equipment Corporation.               */
/* All rights reserved. See copyright notice in attached file COPYRIGHT. */
/* Last modified on Mon Feb 18 15:47:05 1991 by burrows        */
#include <stdio.h>
#include <strings.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <setjmp.h>

extern char *malloc ();
extern char *realloc ();
extern void free ();
extern void abort ();
extern void exit ();
extern void perror ();
extern void bzero ();
extern void bcopy ();
extern long time ();

static void childdeath ();
static int forkserver ();
static void serve ();
static void writedata ();
static void getfromserver ();
static void readdata ();
static void consider ();
static int gchar ();
static void sigpipe ();
static int mywrite ();

static int gpid;

#define MAXFDS	63
#define INTSIZE	32
#define MASKINTS ((MAXFDS+INTSIZE-1) / INTSIZE)

#define MAXDATA	(5*1024)

struct call {
	int flags;
	int maxdata;
	int ndata;
	char *data;
	int maxwdata;
	int nwdata;
	int nWdata;
	char *wdata;
};
#define CALLINUSE	0x0001
#define CALLINPROGRESS	0x0002
#define CALLPENDING	0x0004
#define CALLDEAD	0x0008

static void usage (prog)
char *prog;
{
	if (prog == 0)
		prog = "newsindexserver";

	if (rindex (prog, '/') != 0)
		prog = rindex (prog, '/') + 1;

	(void) fprintf (stderr, "usage: %s [-p port] [-e char] serverprog\n",
		prog);

	exit (2);
}

main (argc, argv)
int argc;
char *argv[];
{
	int argn;
	int i;
	int fd;
	char hostname[128];
	struct sockaddr_in addr;
	struct hostent *hp;
	struct servent *sp;
	char **serverprog;
	int one = 1;
	int endbyte = '\n';
	char *arg;
	char *port = "2314";

#ifdef SETLOCALE
	(void) setlocale (0, "");
#endif

	for (argn = 1; argn < argc; argn++) {
		arg = argv[argn];
		if (arg[0] == '-') {
			for (i = 1; arg[i] != '\0'; i++) {
				switch (arg[i])
				{
				case 'p':
					if (argn+1 >= argc)
						usage (argv[0]);
					port = argv[++argn];
					break;
				case 'e':
					if (argn+1 >= argc)
						usage (argv[0]);
					endbyte = gchar (argv[++argn]);
					break;
				default: usage (argv[0]);
				}
			}
		} else 
			break;
	}

	if (argn >= argc)
		usage (argv[0]);
	serverprog = &argv[argn];

	if ((fd = socket (AF_INET, SOCK_STREAM, 0)) < 0) {
		perror ("socket");
		exit (2);
	}
	if ((i = gethostname (hostname, sizeof (hostname))) < 0 ||
		i == sizeof (hostname)) {
		perror ("gethostname");
		exit (2);
	}
	if ((hp = gethostbyname (hostname)) == 0) {
		(void) fprintf (stderr, "unkonwn host: %s\n", hostname);
		exit (2);
	}

	addr.sin_family = hp->h_addrtype;
	bcopy (hp->h_addr, (char *) &addr.sin_addr, hp->h_length);
	endhostent ();


	if ((sp = getservbyname ("newsindex", "tcp")) == 0) {
		addr.sin_port = htons (atoi (port));
		/*
		(void) fprintf (stderr, "unknown service: %s\n", "newsindex");
		exit (2);
		*/
	} else {
		addr.sin_port = sp->s_port;
	}
	endservent ();

	one = 1;
	if (setsockopt (fd, SOL_SOCKET, SO_REUSEADDR,
			(char *) &one, sizeof (one)) < 0) {
		perror ("SO_REUSEADDR");
	}	
	if (bind (fd, (struct sockaddr *)&addr, sizeof(struct sockaddr_in))<0) {
		perror ("bind");
		exit (2);
	}

	if (listen (fd, 5) < 0) {
		perror ("listen");
		exit (2);
	}

	if (fcntl (fd, F_SETFL, FNDELAY) < 0) {
		perror ("FNDELAY");
		exit (2);
	}

	(void) signal (SIGPIPE, sigpipe);
	(void) signal (SIGCHLD, childdeath);

	serve (fd, serverprog, endbyte);
	/* NOTREACHED */
}

static void childdeath ()
{
	union wait status;
	struct rusage rusage;
	int pid;

	while ((pid = wait3 (&status, WNOHANG, &rusage)) > 0) {
		if (pid == gpid) {
			(void) fprintf (stderr,
				"netserver: child %d died status=0x%x\n",
				    pid, *(int *)&status);
			if (!WIFEXITED (status))
				(void) kill (getpid (), SIGHUP);
		}
	}
}

static int forkserver (prog, pin, pout, closefd)
char **prog;
int pin[];
int pout[];
int closefd;
{
	int pid;

	if (pipe (pin) < 0 || pipe (pout) < 0) {
		perror ("pipe");
		exit (2);
	}

	pid = vfork ();

	if (pid < 0) {
		perror ("vfork");
		exit (2);
	}

	if (pid == 0) { 		/* child */
		if (closefd >= 0)
			(void) close (closefd);
		(void) signal (SIGPIPE, SIG_DFL);
		(void) close (pin[0]);
		(void) close (pout[1]);
		if (pout[0] != 0) {
			(void) dup2 (pout[0], 0);
			(void) close (pout[0]);
		}
		if (pin[1] != 1) {
			(void) dup2 (pin[1], 1);
			(void) close (pin[1]);
		}
		(void) execvp (prog[0], prog);
		perror (prog[0]);
		exit (2);
	}

	(void) close (pin[1]);
	(void) close (pout[0]);

	return (pid);
}

static void serve (afd, serverprog, endbyte)
int afd;
char **serverprog;
int endbyte;
{
	int nfd;
	int i;
	struct sockaddr_in addr;
	int one = 1;
	int len;
	int rmask[MASKINTS], wmask[MASKINTS], emask[MASKINTS];
	int sel;
	struct call calls[MAXFDS];
	int serveroutput = -1;
	struct timeval timeval;
	int pin[2], pout[2];
	int fromserver, toserver;
	int pid;
	int intr = 0;

	pid = forkserver (serverprog, pin, pout, afd);
	gpid = pid;
	fromserver = pin[0];
	toserver = pout[1];

	bzero ((char *) calls, sizeof (calls));

	for (;;) {
		for (i = 0; i < MASKINTS; i++)
			rmask[i] = wmask[i] = 0;
		for (i = 0; i < MAXFDS; i++) {
			if (calls[i].flags & CALLINUSE) {
				if ((calls[i].flags & CALLINPROGRESS) == 0)
					rmask[i / INTSIZE] |= 1 <<(i % INTSIZE);
				else if (calls[i].nwdata > 0)
					wmask[i / INTSIZE] |= 1 <<(i % INTSIZE);
			}
		}
		for (i = 0; i < MASKINTS; i++)
			emask[i] = rmask[i];

		rmask[afd / INTSIZE] |= 1 << (afd % INTSIZE);
		if (serveroutput >= 0)
			rmask[fromserver / INTSIZE] |= 1<<(fromserver%INTSIZE);

		timeval.tv_sec = 60;
		timeval.tv_usec = 0;
		sel = select (MAXFDS, rmask, wmask, emask, &timeval);

		if (sel < 0) {
			if (errno == EINTR)
				continue;
			perror ("select");
			exit (2);
		}

		if (rmask[afd / INTSIZE] & (1 << (afd % INTSIZE))) {
			len = sizeof (struct sockaddr_in);
			if ((nfd = accept (afd, (struct sockaddr *)
					&addr, &len)) >= 0) {
				one = 1;
				if (setsockopt (nfd, SOL_SOCKET, SO_KEEPALIVE, 
					(char *) &one, sizeof (one)) < 0) {
					perror ("SO_KEEPALIVE");
				}	
				if (fcntl (nfd, F_SETFL, FNDELAY) < 0) {
					perror ("FNDELAY");
					exit (2);
				}
			} else if (errno != EINTR || errno != EWOULDBLOCK) {
				perror ("aceept");
				exit (2);
			} else
				perror ("accept");
			if (nfd >= MAXFDS) {
				(void) fprintf (stderr,
					"file descriptor too big\n");
				exit (2);
			}

			if (calls[nfd].flags != 0)
				abort ();

			calls[nfd].flags = CALLINUSE;
			calls[nfd].ndata = 0;
			rmask[afd / INTSIZE] &= ~(1 << (afd % INTSIZE));
		}

		if (rmask[fromserver / INTSIZE] & (1 << (fromserver%INTSIZE))) {
			getfromserver (calls, &serveroutput,
				fromserver, toserver, endbyte, &intr);
			rmask[fromserver / INTSIZE] &=
				~(1 << (fromserver % INTSIZE));
		}

		for (i = 0; i < MASKINTS; i++)
			rmask[i] |= emask[i];

		for (i = 0; i < MAXFDS; i++) {
			if (rmask[i / INTSIZE] & (1 << (i % INTSIZE))) {
				readdata (calls, i, &serveroutput, toserver);
			} else if (wmask[i / INTSIZE] & (1 << (i % INTSIZE))) {
				writedata (calls, i, &serveroutput, toserver,
					endbyte, pid, &intr);
			}
		}
	}
}

static void writedata (calls, fd, serveroutput, toserver, endbyte, pid, intr)
struct call *calls;
int fd;
int *serveroutput;
int toserver;
int endbyte;
int pid;
int *intr;
{
	struct call *call = &calls[fd];
	int wrote;

	if ((call->flags & CALLINUSE) == 0)
		abort ();
	if ((call->flags & CALLINPROGRESS) == 0)
		abort ();

	if ((call->flags & CALLDEAD) == 0) {
		wrote = mywrite (fd, call->wdata+call->nWdata, call->nwdata);
	} else
		wrote = call->nwdata;

	if (wrote < 0) {
		if (errno == EINTR || errno == EWOULDBLOCK) {
			perror ("write");
			return;
		}
		perror ("write to dead client");
		call->flags |= CALLDEAD;
		call->flags &= ~CALLPENDING;
		call->ndata = 0;
		wrote = call->nwdata;
		if (*intr == 0 && *serveroutput == fd) {
			if (kill (pid, SIGINT) < 0) {
				perror ("netserver: kill child");
				exit (2);
			}
			*intr = 1;
		}
	}

	if (wrote == call->nwdata) {
		wrote += call->nWdata;
		call->nwdata = call->nWdata = 0;
		if (call->wdata[wrote-1] == endbyte) {
			call->flags &= ~CALLINPROGRESS;
			call->maxwdata = 0;
			free (call->wdata);
			call->wdata = 0;
			if (call->flags & CALLPENDING)
				consider (call, fd, serveroutput, toserver);
		}
	} else {
		call->nwdata -= wrote;
		call->nWdata += wrote;
		/* bcopy (&call->wdata[wrote], call->wdata, call->nwdata); */
	}
}

static void getfromserver (calls, serveroutput, fromserver, toserver, endbyte, intr)
struct call *calls;
int *serveroutput;
int fromserver;
int toserver;
int endbyte;
int *intr;
{
	int fd = *serveroutput;
	struct call *call = &calls[fd];
	int nread;
	int i;

	if (fd < 0)
		abort ();
	
	if ((call->flags & CALLINUSE) == 0)
		abort ();
	if ((call->flags & CALLINPROGRESS) == 0)
		abort ();
	
	if (call->flags & CALLDEAD)
		call->nwdata = call->nWdata = 0;

	if (call->nWdata+call->nwdata >= call->maxwdata) {
		call->maxwdata = (call->nWdata+call->nwdata+0x3fff) & ~0x3ff;
		if ((call->wdata = realloc (call->wdata,
				(unsigned) call->maxwdata))==0) {
			perror ("realloc");
			exit (2);
		}
	}

	nread = read (fromserver, call->wdata+call->nWdata+call->nwdata,
		call->maxwdata-call->nWdata-call->nwdata);

	if (nread > 0) {
		call->nwdata += nread;
		if (*intr == 1 && call->nwdata > 0 &&
		       call->wdata[call->nWdata+call->nwdata-1] == endbyte) {
			call->nwdata--;
			*intr = 2;
			do {
				i = write (toserver, "\n", 1);
			} while (i < 0 && errno == EINTR);
			if (i < 0) {
				perror ("write to server");
				exit (2);
			}
		}
		if (call->nwdata > 0 && call->wdata[call->nWdata+call->nwdata-1]==endbyte) {
			*serveroutput = -1;
			*intr = 0;
			if (call->flags & CALLDEAD) {
				call->ndata = 0;
				call->maxdata = 0;
				free (call->data);
				call->data = 0;
				call->nwdata = 0;
				call->nWdata = 0;
				call->maxwdata = 0;
				free (call->wdata);
				call->wdata = 0;
				call->flags = 0;
				(void) close (fd);
			} else {
				for (i = fd+1; i < MAXFDS && *serveroutput < 0; i++)
					if (calls[i].flags & CALLPENDING)
						consider (&calls[i], i,
							serveroutput, toserver);
				for (i = 0; i < fd && *serveroutput < 0; i++) 
					if (calls[i].flags & CALLPENDING)
						consider (&calls[i], i,
							serveroutput, toserver);
			}
		} else if (call->flags & CALLDEAD)
			call->nwdata = call->nWdata = 0;

		return;
	}

	if (nread < 0) {
		if (errno == EINTR || errno == EWOULDBLOCK) {
			perror ("read from server");
			return;
		}
		perror ("read from server");
	} else
		(void) fprintf (stderr, "EOF fromserver\n");

	(void) kill (getpid (), SIGHUP);

	exit (2);
}


static void readdata (calls, fd, serveroutput, toserver)
struct call *calls;
int fd;
int *serveroutput;
int toserver;
{
	struct call *call = &calls[fd];
	int nread;

	if ((call->flags & CALLINUSE) == 0)
		abort ();
	if ((call->flags & CALLINPROGRESS) != 0)
		abort ();

	if (call->ndata >= call->maxdata || (call->flags & CALLDEAD)) {
		if (call->ndata > MAXDATA || (call->flags & CALLDEAD)) {
			call->ndata = 0;
			call->maxdata = 0;
			free (call->data);
			call->data = 0;
			call->nwdata = 0;
			call->nWdata = 0;
			call->maxwdata = 0;
			free (call->wdata);
			call->wdata = 0;
			call->flags = 0;
			(void) close (fd);
			return;
		}
		call->maxdata = (call->ndata+0x7ff) & ~0x3ff;
		if ((call->data = realloc (call->data,
				(unsigned) call->maxdata)) == 0) {
			perror ("realloc");
			exit (2);
		}
	}

	nread = read (fd, call->data+call->ndata, call->maxdata-call->ndata);

	if (nread > 0) {
		call->ndata += nread;
		consider (call, fd, serveroutput, toserver);
		return;
	}
	if (nread < 0 && (errno == EINTR || errno == EWOULDBLOCK)) {
		perror ("read from client");
		return;
	}

	if (nread < 0)
		perror ("read from dead client");

	call->ndata = 0;
	call->maxdata = 0;
	free (call->data);
	call->data = 0;
	call->nwdata = 0;
	call->nWdata = 0;
	call->maxwdata = 0;
	free (call->wdata);
	call->wdata = 0;
	call->flags = 0;
	(void) close (fd);
	return;
}

static void consider (call, fd, serveroutput, toserver)
struct call *call;
int fd;
int *serveroutput;
int toserver;
{
	int done;
	int wrote;
	int i = 0;

	if (call->flags & CALLINPROGRESS)
		return;

	if ((call->flags & CALLPENDING) && (*serveroutput >= 0))
		return;

	while (i < call->ndata && call->data[i] != '\n')
		i++;
	if (i >= call->ndata) {
		call->flags &= ~CALLPENDING;
		return;
	}
	
	if (*serveroutput >= 0) {
		call->flags |= CALLPENDING;
		return;
	}

	i++;
	done = 0;
	while (done < i) {
		wrote = mywrite (toserver, call->data+done, i - done);
		if (wrote > 0)
			done += wrote;
		else if (wrote < 0 && errno != EINTR) {
			perror ("write to server");
			exit (2);
		}
	}

	if (i == call->ndata)
		call->ndata = 0;
	else {
		call->ndata -= i;
		bcopy (&call->data[i], call->data, call->ndata);
	}
	call->flags |= CALLINPROGRESS;
	*serveroutput = fd;
}

static int gchar (s)
char *s;
{
	int x;
	int i;
	static char hex[] = "0123456789abcdef";
	char *p;
	int c;

	if (*s == '\\' && s[1] != '\0') {
		s++;
		if (isdigit (*s) && *s != '8' && *s != '9') {
			x = 0;
			i = 0;
			while (i<3 && isdigit (*s) && *s != '8' && *s != '9') {
				x <<= 3;
				x |= index (hex, *s++) - hex;
				i++;
			}
			return (x);
		} else {
			switch (*s)
			{
			case 'x':
				s++;
				x = 0;
				i = 0;
				do {
					if ((c = *s++) != 0) {
						p = index (hex, tolower (c));
						if (p != 0) {
							x <<= 4;
							x |= p - hex;
						}
						i++;
					}
				} while (i < 2 && c != 0 && p != 0);
				return (x);
			case 't': return ('\t');
			case 'b': return ('\b');
			case 'v': return ('\v');
			case 's': return (' ');
			case 'r': return ('\r');
			case 'e': return ('\033');
			case 'n': return ('\n');
			default:  return (*s);
			}
		}
	} else
		return (*s);
}

static int sigpipereturn = 0;
jmp_buf sigpipejmp;

static void sigpipe ()
{
	if (sigpipereturn)
		longjmp (sigpipejmp, 1);
}

static int mywrite (fd, buf, len)
int fd;
char *buf;
int len;
{
	int n;

	if (setjmp (sigpipejmp) == 0) {
		sigpipereturn = 1;
		n = write (fd, buf, len);
		sigpipereturn = 0;
	} else {
		errno = EPIPE;
		return (-1);
	}
	return (n);
}
